from struct import Struct


def read_binary_str(buf):
    length = read_varint(buf)
    return buf.read(length).decode('utf-8')


def read_binary_str_fixed_len(buf, length):
    return buf.read(length).decode('utf-8')


def _read_one(f):
    c = f.read(1)
    if c == b'':
        raise EOFError("Unexpected EOF while reading bytes")

    return ord(c)


def read_varint(f):
    """
    Reads integer of variable length using LEB128.
    """
    shift = 0
    result = 0

    while True:
        i = _read_one(f)
        result |= (i & 0x7f) << shift
        shift += 7
        if not (i & 0x80):
            break

    return result


def read_binary_int(buf, fmt):
    """
    Reads int from buffer with provided format.
    """
    # Little endian.
    s = Struct('<' + fmt)
    return s.unpack(buf.read(s.size))[0]


def read_binary_int8(buf):
    return read_binary_int(buf, 'b')


def read_binary_int16(buf):
    return read_binary_int(buf, 'h')


def read_binary_int32(buf):
    return read_binary_int(buf, 'i')


def read_binary_int64(buf):
    return read_binary_int(buf, 'q')


def read_binary_uint8(buf):
    return read_binary_int(buf, 'B')


def read_binary_uint16(buf):
    return read_binary_int(buf, 'H')


def read_binary_uint32(buf):
    return read_binary_int(buf, 'I')


def read_binary_uint64(buf):
    return read_binary_int(buf, 'Q')


def read_binary_uint128(buf):
    hi = read_binary_int(buf, 'Q')
    lo = read_binary_int(buf, 'Q')

    return (hi << 64) + lo



from ctypes import c_float

from .base import FormatColumn


class FloatColumn(FormatColumn):
    py_types = (float, int)


class Float32(FloatColumn):
    ch_type = 'Float32'
    format = 'f'

    def __init__(self, types_check=False, **kwargs):
        super(Float32, self).__init__(types_check=types_check, **kwargs)

        if types_check:
            # Chop only bytes that fit current type.
            # Cast to -nan or nan if overflows.
            def before_write_item(value):
                return c_float(value).value

            self.before_write_item = before_write_item


class Float64(FloatColumn):
    ch_type = 'Float64'
    format = 'd'



import socket
import getpass

from . import defines
from . import errors
from .writer import write_binary_str, write_varint, write_binary_uint8


class ClientInfo(object):
    class Interface(object):
        TCP = 1
        HTTP = 2

    class QueryKind(object):
        # Uninitialized object.
        NO_QUERY = 0

        INITIAL_QUERY = 1

        # Query that was initiated by another query for distributed query
        # execution.
        SECONDARY_QUERY = 2

    client_version_major = defines.DBMS_VERSION_MAJOR
    client_version_minor = defines.DBMS_VERSION_MINOR
    client_revision = defines.CLIENT_VERSION
    interface = Interface.TCP

    initial_user = ''
    initial_query_id = ''
    initial_address = '0.0.0.0:0'

    quota_key = ''

    def __init__(self, client_name):
        self.query_kind = ClientInfo.QueryKind.NO_QUERY

        self.os_user = getpass.getuser()
        self.client_hostname = socket.gethostname()
        self.client_name = client_name

        super(ClientInfo, self).__init__()

    @property
    def empty(self):
        return self.query_kind == ClientInfo.QueryKind.NO_QUERY

    def write(self, server_revision, fout):
        revision = server_revision
        if server_revision < defines.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
            raise errors.LogicalError('Method ClientInfo.write is called '
                                      'for unsupported server revision')

        write_binary_uint8(self.query_kind, fout)
        if self.empty:
            return

        write_binary_str(self.initial_user, fout)
        write_binary_str(self.initial_query_id, fout)
        write_binary_str(self.initial_address, fout)

        write_binary_uint8(self.interface, fout)

        write_binary_str(self.os_user, fout)
        write_binary_str(self.client_hostname, fout)
        write_binary_str(self.client_name, fout)
        write_varint(self.client_version_major, fout)
        write_varint(self.client_version_minor, fout)
        write_varint(self.client_revision, fout)

        if revision >= defines.DBMS_MIN_REVISION_WITH_QUOTA_KEY_IN_CLIENT_INFO:
            write_binary_str(self.quota_key, fout)



from .intcolumn import FormatColumn


# TODO: Drop Null column support in future.
# Compatibility with old servers.
class NullColumn(FormatColumn):
    ch_type = 'Null'
    format = 'B'

    @property
    def size(self):
        return 1

    def after_read_item(self, buf):
        return None



from . import defines
from .reader import read_varint


class Progress(object):
    def __init__(self):
        self.new_rows = 0
        self.new_bytes = 0
        self.new_total_rows = 0

        super(Progress, self).__init__()

    def read(self, server_revision, fin):
        self.new_rows = read_varint(fin)
        self.new_bytes = read_varint(fin)

        revision = server_revision
        if revision >= defines.DBMS_MIN_REVISION_WITH_TOTAL_ROWS_IN_PROGRESS:
            self.new_total_rows = read_varint(fin)



from struct import Struct, error as struct_error

from . import exceptions


class Column(object):
    ch_type = None
    py_types = None

    check_item = None
    after_read_item = None
    before_write_item = None

    types_check_enabled = False

    def __init__(self, types_check=False, **kwargs):
        self.nullable = False
        self.types_check_enabled = types_check
        super(Column, self).__init__()

    def make_null_struct(self, n_items):
        return Struct('<{}B'.format(n_items))

    def _read_nulls_map(self, n_items, buf):
        s = self.make_null_struct(n_items)
        return s.unpack(buf.read(s.size))

    def _write_nulls_map(self, items, buf):
        s = self.make_null_struct(len(items))
        items = [x is None for x in items]
        buf.write(s.pack(*items))

    def prepare_null(self, value):
        if self.nullable and value is None:
            return 0, True

        else:
            return value, False

    def check_item_type(self, value):
        if not isinstance(value, self.py_types):
            raise exceptions.ColumnTypeMismatchException(value)

    def prepare_items(self, items):
        before_write = self.before_write_item
        prepare_null = self.prepare_null if self.nullable else False

        check_item = self.check_item
        if self.types_check_enabled:
            check_item_type = self.check_item_type
        else:
            check_item_type = False

        prepared = [None] * len(items)
        for i, x in enumerate(items):
            if prepare_null:
                x, is_null = prepare_null(x)
            else:
                is_null = False

            if not is_null:
                if check_item_type:
                    check_item_type(x)

                if check_item:
                    check_item(x)

                if before_write:
                    x = before_write(x)

            prepared[i] = x

        return prepared

    def write_data(self, items, buf):
        if self.nullable:
            self._write_nulls_map(items, buf)

        self._write_data(items, buf)

    def _write_data(self, items, buf):
        prepared = self.prepare_items(items)
        self.write_items(prepared, buf)

    def write_items(self, items, buf):
        raise NotImplementedError

    def read_data(self, n_items, buf):
        if self.nullable:
            nulls_map = self._read_nulls_map(n_items, buf)
        else:
            nulls_map = None

        return self._read_data(n_items, buf, nulls_map=nulls_map)

    def _read_data(self, n_items, buf, nulls_map=None):
        after_read = self.after_read_item

        items = self.read_items(n_items, buf)

        if nulls_map is not None:
            if after_read:
                items = tuple(
                    (None if is_null else after_read(items[i]))
                    for i, is_null in enumerate(nulls_map)
                )
            else:
                items = tuple(
                    (None if is_null else items[i])
                    for i, is_null in enumerate(nulls_map)
                )

        else:
            if after_read:
                items = (after_read(x) for x in items)

        return tuple(items)

    def read_items(self, n_items, buf):
        raise NotImplementedError


class FormatColumn(Column):
    """
    Uses struct.pack for bulk items writing.
    """

    format = None

    def make_struct(self, n_items):
        return Struct('<{}{}'.format(n_items, self.format))

    def write_items(self, items, buf):
        s = self.make_struct(len(items))
        try:
            buf.write(s.pack(*items))

        except struct_error as e:
            raise exceptions.StructPackException(e)

    def read_items(self, n_items, buf):
        s = self.make_struct(n_items)
        return s.unpack(buf.read(s.size))


class CustomItemColumn(Column):
    """
    Providers per item read/write functions.
    This column should be used when column can't be bulk-processed.
    """

    def read(self, buf):
        raise NotImplementedError

    def _read_null(self, buf):
        raise NotImplementedError

    def write(self, value, buf):
        raise NotImplementedError

    def _write_null(self, buf):
        raise NotImplementedError

    def _write_data(self, items, buf):
        if self.types_check_enabled:
            check_item_type = self.check_item_type
        else:
            check_item_type = False

        for x in items:
            self.write_item(x, buf, check_item_type)

    def write_item(self, x, buf, check_item_type):
        if self.nullable and x is None:
            self._write_null(buf)

        else:
            if check_item_type:
                check_item_type(x)

            self.write(x, buf)

    def _read_data(self, n_items, buf, nulls_map=None):
        if nulls_map is not None:
            return tuple(
                self.read_item(buf, is_null=is_null) for is_null in nulls_map
            )

        else:
            return tuple(self.read(buf) for i in range(n_items))

    def read_item(self, buf, is_null=False):
        return self._read_null(buf) if is_null else self.read(buf)




from .util import compat


class ErrorCodes(object):
    UNSUPPORTED_METHOD = 1
    UNSUPPORTED_PARAMETER = 2
    UNEXPECTED_END_OF_FILE = 3
    EXPECTED_END_OF_FILE = 4
    CANNOT_PARSE_TEXT = 6
    INCORRECT_NUMBER_OF_COLUMNS = 7
    THERE_IS_NO_COLUMN = 8
    SIZES_OF_COLUMNS_DOESNT_MATCH = 9
    NOT_FOUND_COLUMN_IN_BLOCK = 10
    POSITION_OUT_OF_BOUND = 11
    PARAMETER_OUT_OF_BOUND = 12
    SIZES_OF_COLUMNS_IN_TUPLE_DOESNT_MATCH = 13
    DUPLICATE_COLUMN = 15
    NO_SUCH_COLUMN_IN_TABLE = 16
    DELIMITER_IN_STRING_LITERAL_DOESNT_MATCH = 17
    CANNOT_INSERT_ELEMENT_INTO_CONSTANT_COLUMN = 18
    SIZE_OF_FIXED_STRING_DOESNT_MATCH = 19
    NUMBER_OF_COLUMNS_DOESNT_MATCH = 20
    CANNOT_READ_ALL_DATA_FROM_TAB_SEPARATED_INPUT = 21
    CANNOT_PARSE_ALL_VALUE_FROM_TAB_SEPARATED_INPUT = 22
    CANNOT_READ_FROM_ISTREAM = 23
    CANNOT_WRITE_TO_OSTREAM = 24
    CANNOT_PARSE_ESCAPE_SEQUENCE = 25
    CANNOT_PARSE_QUOTED_STRING = 26
    CANNOT_PARSE_INPUT_ASSERTION_FAILED = 27
    CANNOT_PRINT_FLOAT_OR_DOUBLE_NUMBER = 28
    CANNOT_PRINT_INTEGER = 29
    CANNOT_READ_SIZE_OF_COMPRESSED_CHUNK = 30
    CANNOT_READ_COMPRESSED_CHUNK = 31
    ATTEMPT_TO_READ_AFTER_EOF = 32
    CANNOT_READ_ALL_DATA = 33
    TOO_MUCH_ARGUMENTS_FOR_FUNCTION = 34
    TOO_LESS_ARGUMENTS_FOR_FUNCTION = 35
    BAD_ARGUMENTS = 36
    UNKNOWN_ELEMENT_IN_AST = 37
    CANNOT_PARSE_DATE = 38
    TOO_LARGE_SIZE_COMPRESSED = 39
    CHECKSUM_DOESNT_MATCH = 40
    CANNOT_PARSE_DATETIME = 41
    NUMBER_OF_ARGUMENTS_DOESNT_MATCH = 42
    ILLEGAL_TYPE_OF_ARGUMENT = 43
    ILLEGAL_COLUMN = 44
    ILLEGAL_NUMBER_OF_RESULT_COLUMNS = 45
    UNKNOWN_FUNCTION = 46
    UNKNOWN_IDENTIFIER = 47
    NOT_IMPLEMENTED = 48
    LOGICAL_ERROR = 49
    UNKNOWN_TYPE = 50
    EMPTY_LIST_OF_COLUMNS_QUERIED = 51
    COLUMN_QUERIED_MORE_THAN_ONCE = 52
    TYPE_MISMATCH = 53
    STORAGE_DOESNT_ALLOW_PARAMETERS = 54
    STORAGE_REQUIRES_PARAMETER = 55
    UNKNOWN_STORAGE = 56
    TABLE_ALREADY_EXISTS = 57
    TABLE_METADATA_ALREADY_EXISTS = 58
    ILLEGAL_TYPE_OF_COLUMN_FOR_FILTER = 59
    UNKNOWN_TABLE = 60
    ONLY_FILTER_COLUMN_IN_BLOCK = 61
    SYNTAX_ERROR = 62
    UNKNOWN_AGGREGATE_FUNCTION = 63
    CANNOT_READ_AGGREGATE_FUNCTION_FROM_TEXT = 64
    CANNOT_WRITE_AGGREGATE_FUNCTION_AS_TEXT = 65
    NOT_A_COLUMN = 66
    ILLEGAL_KEY_OF_AGGREGATION = 67
    CANNOT_GET_SIZE_OF_FIELD = 68
    ARGUMENT_OUT_OF_BOUND = 69
    CANNOT_CONVERT_TYPE = 70
    CANNOT_WRITE_AFTER_END_OF_BUFFER = 71
    CANNOT_PARSE_NUMBER = 72
    UNKNOWN_FORMAT = 73
    CANNOT_READ_FROM_FILE_DESCRIPTOR = 74
    CANNOT_WRITE_TO_FILE_DESCRIPTOR = 75
    CANNOT_OPEN_FILE = 76
    CANNOT_CLOSE_FILE = 77
    UNKNOWN_TYPE_OF_QUERY = 78
    INCORRECT_FILE_NAME = 79
    INCORRECT_QUERY = 80
    UNKNOWN_DATABASE = 81
    DATABASE_ALREADY_EXISTS = 82
    DIRECTORY_DOESNT_EXIST = 83
    DIRECTORY_ALREADY_EXISTS = 84
    FORMAT_IS_NOT_SUITABLE_FOR_INPUT = 85
    RECEIVED_ERROR_FROM_REMOTE_IO_SERVER = 86
    CANNOT_SEEK_THROUGH_FILE = 87
    CANNOT_TRUNCATE_FILE = 88
    UNKNOWN_COMPRESSION_METHOD = 89
    EMPTY_LIST_OF_COLUMNS_PASSED = 90
    SIZES_OF_MARKS_FILES_ARE_INCONSISTENT = 91
    EMPTY_DATA_PASSED = 92
    UNKNOWN_AGGREGATED_DATA_VARIANT = 93
    CANNOT_MERGE_DIFFERENT_AGGREGATED_DATA_VARIANTS = 94
    CANNOT_READ_FROM_SOCKET = 95
    CANNOT_WRITE_TO_SOCKET = 96
    CANNOT_READ_ALL_DATA_FROM_CHUNKED_INPUT = 97
    CANNOT_WRITE_TO_EMPTY_BLOCK_OUTPUT_STREAM = 98
    UNKNOWN_PACKET_FROM_CLIENT = 99
    UNKNOWN_PACKET_FROM_SERVER = 100
    UNEXPECTED_PACKET_FROM_CLIENT = 101
    UNEXPECTED_PACKET_FROM_SERVER = 102
    RECEIVED_DATA_FOR_WRONG_QUERY_ID = 103
    TOO_SMALL_BUFFER_SIZE = 104
    CANNOT_READ_HISTORY = 105
    CANNOT_APPEND_HISTORY = 106
    FILE_DOESNT_EXIST = 107
    NO_DATA_TO_INSERT = 108
    CANNOT_BLOCK_SIGNAL = 109
    CANNOT_UNBLOCK_SIGNAL = 110
    CANNOT_MANIPULATE_SIGSET = 111
    CANNOT_WAIT_FOR_SIGNAL = 112
    THERE_IS_NO_SESSION = 113
    CANNOT_CLOCK_GETTIME = 114
    UNKNOWN_SETTING = 115
    THERE_IS_NO_DEFAULT_VALUE = 116
    INCORRECT_DATA = 117
    TABLE_METADATA_DOESNT_EXIST = 118
    ENGINE_REQUIRED = 119
    CANNOT_INSERT_VALUE_OF_DIFFERENT_SIZE_INTO_TUPLE = 120
    UNKNOWN_SET_DATA_VARIANT = 121
    INCOMPATIBLE_COLUMNS = 122
    UNKNOWN_TYPE_OF_AST_NODE = 123
    INCORRECT_ELEMENT_OF_SET = 124
    INCORRECT_RESULT_OF_SCALAR_SUBQUERY = 125
    CANNOT_GET_RETURN_TYPE = 126
    ILLEGAL_INDEX = 127
    TOO_LARGE_ARRAY_SIZE = 128
    FUNCTION_IS_SPECIAL = 129
    CANNOT_READ_ARRAY_FROM_TEXT = 130
    TOO_LARGE_STRING_SIZE = 131
    CANNOT_CREATE_TABLE_FROM_METADATA = 132
    AGGREGATE_FUNCTION_DOESNT_ALLOW_PARAMETERS = 133
    PARAMETERS_TO_AGGREGATE_FUNCTIONS_MUST_BE_LITERALS = 134
    ZERO_ARRAY_OR_TUPLE_INDEX = 135
    UNKNOWN_ELEMENT_IN_CONFIG = 137
    EXCESSIVE_ELEMENT_IN_CONFIG = 138
    NO_ELEMENTS_IN_CONFIG = 139
    ALL_REQUESTED_COLUMNS_ARE_MISSING = 140
    SAMPLING_NOT_SUPPORTED = 141
    NOT_FOUND_NODE = 142
    FOUND_MORE_THAN_ONE_NODE = 143
    FIRST_DATE_IS_BIGGER_THAN_LAST_DATE = 144
    UNKNOWN_OVERFLOW_MODE = 145
    QUERY_SECTION_DOESNT_MAKE_SENSE = 146
    NOT_FOUND_FUNCTION_ELEMENT_FOR_AGGREGATE = 147
    NOT_FOUND_RELATION_ELEMENT_FOR_CONDITION = 148
    NOT_FOUND_RHS_ELEMENT_FOR_CONDITION = 149
    NO_ATTRIBUTES_LISTED = 150
    INDEX_OF_COLUMN_IN_SORT_CLAUSE_IS_OUT_OF_RANGE = 151
    UNKNOWN_DIRECTION_OF_SORTING = 152
    ILLEGAL_DIVISION = 153
    AGGREGATE_FUNCTION_NOT_APPLICABLE = 154
    UNKNOWN_RELATION = 155
    DICTIONARIES_WAS_NOT_LOADED = 156
    ILLEGAL_OVERFLOW_MODE = 157
    TOO_MUCH_ROWS = 158
    TIMEOUT_EXCEEDED = 159
    TOO_SLOW = 160
    TOO_MUCH_COLUMNS = 161
    TOO_DEEP_SUBQUERIES = 162
    TOO_DEEP_PIPELINE = 163
    READONLY = 164
    TOO_MUCH_TEMPORARY_COLUMNS = 165
    TOO_MUCH_TEMPORARY_NON_CONST_COLUMNS = 166
    TOO_DEEP_AST = 167
    TOO_BIG_AST = 168
    BAD_TYPE_OF_FIELD = 169
    BAD_GET = 170
    BLOCKS_HAVE_DIFFERENT_STRUCTURE = 171
    CANNOT_CREATE_DIRECTORY = 172
    CANNOT_ALLOCATE_MEMORY = 173
    CYCLIC_ALIASES = 174
    CHUNK_NOT_FOUND = 176
    DUPLICATE_CHUNK_NAME = 177
    MULTIPLE_ALIASES_FOR_EXPRESSION = 178
    MULTIPLE_EXPRESSIONS_FOR_ALIAS = 179
    THERE_IS_NO_PROFILE = 180
    ILLEGAL_FINAL = 181
    ILLEGAL_PREWHERE = 182
    UNEXPECTED_EXPRESSION = 183
    ILLEGAL_AGGREGATION = 184
    UNSUPPORTED_MYISAM_BLOCK_TYPE = 185
    UNSUPPORTED_COLLATION_LOCALE = 186
    COLLATION_COMPARISON_FAILED = 187
    UNKNOWN_ACTION = 188
    TABLE_MUST_NOT_BE_CREATED_MANUALLY = 189
    SIZES_OF_ARRAYS_DOESNT_MATCH = 190
    SET_SIZE_LIMIT_EXCEEDED = 191
    UNKNOWN_USER = 192
    WRONG_PASSWORD = 193
    REQUIRED_PASSWORD = 194
    IP_ADDRESS_NOT_ALLOWED = 195
    UNKNOWN_ADDRESS_PATTERN_TYPE = 196
    SERVER_REVISION_IS_TOO_OLD = 197
    DNS_ERROR = 198
    UNKNOWN_QUOTA = 199
    QUOTA_DOESNT_ALLOW_KEYS = 200
    QUOTA_EXPIRED = 201
    TOO_MUCH_SIMULTANEOUS_QUERIES = 202
    NO_FREE_CONNECTION = 203
    CANNOT_FSYNC = 204
    NESTED_TYPE_TOO_DEEP = 205
    ALIAS_REQUIRED = 206
    AMBIGUOUS_IDENTIFIER = 207
    EMPTY_NESTED_TABLE = 208
    SOCKET_TIMEOUT = 209
    NETWORK_ERROR = 210
    EMPTY_QUERY = 211
    UNKNOWN_LOAD_BALANCING = 212
    UNKNOWN_TOTALS_MODE = 213
    CANNOT_STATVFS = 214
    NOT_AN_AGGREGATE = 215
    QUERY_WITH_SAME_ID_IS_ALREADY_RUNNING = 216
    CLIENT_HAS_CONNECTED_TO_WRONG_PORT = 217
    TABLE_IS_DROPPED = 218
    DATABASE_NOT_EMPTY = 219
    DUPLICATE_INTERSERVER_IO_ENDPOINT = 220
    NO_SUCH_INTERSERVER_IO_ENDPOINT = 221
    ADDING_REPLICA_TO_NON_EMPTY_TABLE = 222
    UNEXPECTED_AST_STRUCTURE = 223
    REPLICA_IS_ALREADY_ACTIVE = 224
    NO_ZOOKEEPER = 225
    NO_FILE_IN_DATA_PART = 226
    UNEXPECTED_FILE_IN_DATA_PART = 227
    BAD_SIZE_OF_FILE_IN_DATA_PART = 228
    QUERY_IS_TOO_LARGE = 229
    NOT_FOUND_EXPECTED_DATA_PART = 230
    TOO_MANY_UNEXPECTED_DATA_PARTS = 231
    NO_SUCH_DATA_PART = 232
    BAD_DATA_PART_NAME = 233
    NO_REPLICA_HAS_PART = 234
    DUPLICATE_DATA_PART = 235
    ABORTED = 236
    NO_REPLICA_NAME_GIVEN = 237
    FORMAT_VERSION_TOO_OLD = 238
    CANNOT_MUNMAP = 239
    CANNOT_MREMAP = 240
    MEMORY_LIMIT_EXCEEDED = 241
    TABLE_IS_READ_ONLY = 242
    NOT_ENOUGH_SPACE = 243
    UNEXPECTED_ZOOKEEPER_ERROR = 244
    INVALID_NESTED_NAME = 245
    CORRUPTED_DATA = 246
    INCORRECT_MARK = 247
    INVALID_PARTITION_VALUE = 248
    NOT_ENOUGH_BLOCK_NUMBERS = 250
    NO_SUCH_REPLICA = 251
    TOO_MUCH_PARTS = 252
    REPLICA_IS_ALREADY_EXIST = 253
    NO_ACTIVE_REPLICAS = 254
    TOO_MUCH_RETRIES_TO_FETCH_PARTS = 255
    PARTITION_ALREADY_EXISTS = 256
    PARTITION_DOESNT_EXIST = 257
    UNION_ALL_RESULT_STRUCTURES_MISMATCH = 258
    UNION_ALL_COLUMN_ALIAS_MISMATCH = 259
    CLIENT_OUTPUT_FORMAT_SPECIFIED = 260
    UNKNOWN_BLOCK_INFO_FIELD = 261
    BAD_COLLATION = 262
    CANNOT_COMPILE_CODE = 263
    INCOMPATIBLE_TYPE_OF_JOIN = 264
    NO_AVAILABLE_REPLICA = 265
    MISMATCH_REPLICAS_DATA_SOURCES = 266
    STORAGE_DOESNT_SUPPORT_PARALLEL_REPLICAS = 267
    CPUID_ERROR = 268
    INFINITE_LOOP = 269
    CANNOT_COMPRESS = 270
    CANNOT_DECOMPRESS = 271
    AIO_SUBMIT_ERROR = 272
    AIO_COMPLETION_ERROR = 273
    AIO_READ_ERROR = 274
    AIO_WRITE_ERROR = 275
    INDEX_NOT_USED = 277
    LEADERSHIP_LOST = 278
    ALL_CONNECTION_TRIES_FAILED = 279
    NO_AVAILABLE_DATA = 280
    DICTIONARY_IS_EMPTY = 281
    INCORRECT_INDEX = 282
    UNKNOWN_DISTRIBUTED_PRODUCT_MODE = 283
    UNKNOWN_GLOBAL_SUBQUERIES_METHOD = 284
    TOO_LESS_LIVE_REPLICAS = 285
    UNSATISFIED_QUORUM_FOR_PREVIOUS_WRITE = 286
    UNKNOWN_FORMAT_VERSION = 287
    DISTRIBUTED_IN_JOIN_SUBQUERY_DENIED = 288
    REPLICA_IS_NOT_IN_QUORUM = 289
    LIMIT_EXCEEDED = 290
    DATABASE_ACCESS_DENIED = 291
    LEADERSHIP_CHANGED = 292
    MONGODB_CANNOT_AUTHENTICATE = 293
    INVALID_BLOCK_EXTRA_INFO = 294
    RECEIVED_EMPTY_DATA = 295
    NO_REMOTE_SHARD_FOUND = 296
    SHARD_HAS_NO_CONNECTIONS = 297
    CANNOT_PIPE = 298
    CANNOT_FORK = 299
    CANNOT_DLSYM = 300
    CANNOT_CREATE_CHILD_PROCESS = 301
    CHILD_WAS_NOT_EXITED_NORMALLY = 302
    CANNOT_SELECT = 303
    CANNOT_WAITPID = 304
    TABLE_WAS_NOT_DROPPED = 305
    TOO_DEEP_RECURSION = 306
    TOO_MUCH_BYTES = 307
    UNEXPECTED_NODE_IN_ZOOKEEPER = 308
    FUNCTION_CANNOT_HAVE_PARAMETERS = 309
    INCONSISTENT_TABLE_ACCROSS_SHARDS = 310
    INSUFFICIENT_SPACE_FOR_RESHARDING = 311
    PARTITION_COPY_FAILED = 312
    PARTITION_ATTACH_FAILED = 313
    RESHARDING_NO_WORKER = 314
    INVALID_PARTITIONS_INTERVAL = 315
    RESHARDING_INVALID_PARAMETERS = 316
    INVALID_SHARD_WEIGHT = 317
    INVALID_CONFIG_PARAMETER = 318
    UNKNOWN_STATUS_OF_INSERT = 319
    DUPLICATE_SHARD_PATHS = 320
    VALUE_IS_OUT_OF_RANGE_OF_DATA_TYPE = 321
    RESHARDING_BUSY_CLUSTER = 322
    RESHARDING_BUSY_SHARD = 323
    RESHARDING_NO_SUCH_COORDINATOR = 324
    RESHARDING_NO_COORDINATOR_MEMBERSHIP = 325
    RESHARDING_ALREADY_SUBSCRIBED = 326
    RESHARDING_REMOTE_NODE_UNAVAILABLE = 327
    RESHARDING_REMOTE_NODE_ERROR = 328
    RESHARDING_COORDINATOR_DELETED = 329
    RESHARDING_DISTRIBUTED_JOB_ON_HOLD = 330
    RESHARDING_INVALID_QUERY = 331
    RESHARDING_INITIATOR_CHECK_FAILED = 332
    RWLOCK_ALREADY_HELD = 333
    RWLOCK_NO_SUCH_LOCK = 334
    BARRIER_TIMEOUT = 335
    UNKNOWN_DATABASE_ENGINE = 336
    DDL_GUARD_IS_ACTIVE = 337
    NO_SUCH_BARRIER = 338
    RESHARDING_ILL_FORMED_LOG = 339
    NO_ZOOKEEPER_ACCESSOR = 340
    UNFINISHED = 341
    METADATA_MISMATCH = 342
    SUPPORT_IS_DISABLED = 344
    TABLE_DIFFERS_TOO_MUCH = 345
    CANNOT_CONVERT_CHARSET = 346
    CANNOT_LOAD_CONFIG = 347
    RESHARDING_NULLABLE_SHARDING_KEY = 348
    CANNOT_INSERT_NULL_IN_ORDINARY_COLUMN = 349
    INCOMPATIBLE_SOURCE_TABLES = 350
    AMBIGUOUS_TABLE_NAME = 351
    AMBIGUOUS_COLUMN_NAME = 352
    INDEX_OF_POSITIONAL_ARGUMENT_IS_OUT_OF_RANGE = 353
    ZLIB_INFLATE_FAILED = 354
    ZLIB_DEFLATE_FAILED = 355
    BAD_LAMBDA = 356
    RESERVED_IDENTIFIER_NAME = 357
    INTO_OUTFILE_NOT_ALLOWED = 358
    TABLE_SIZE_EXCEEDS_MAX_DROP_SIZE_LIMIT = 359
    CANNOT_CREATE_CHARSET_CONVERTER = 360
    SEEK_POSITION_OUT_OF_BOUND = 361
    CURRENT_WRITE_BUFFER_IS_EXHAUSTED = 362
    CANNOT_CREATE_IO_BUFFER = 363
    RECEIVED_ERROR_TOO_MANY_REQUESTS = 364
    OUTPUT_IS_NOT_SORTED = 365
    SIZES_OF_NESTED_COLUMNS_ARE_INCONSISTENT = 366
    TOO_MUCH_FETCHES = 367
    BAD_CAST = 368
    ALL_REPLICAS_ARE_STALE = 369
    DATA_TYPE_CANNOT_BE_USED_IN_TABLES = 370
    INCONSISTENT_CLUSTER_DEFINITION = 371
    SESSION_NOT_FOUND = 372
    SESSION_IS_LOCKED = 373
    INVALID_SESSION_TIMEOUT = 374
    CANNOT_DLOPEN = 375
    CANNOT_PARSE_UUID = 376
    ILLEGAL_SYNTAX_FOR_DATA_TYPE = 377
    DATA_TYPE_CANNOT_HAVE_ARGUMENTS = 378
    UNKNOWN_STATUS_OF_DISTRIBUTED_DDL_TASK = 379
    CANNOT_KILL = 380
    HTTP_LENGTH_REQUIRED = 381

    KEEPER_EXCEPTION = 999
    POCO_EXCEPTION = 1000
    STD_EXCEPTION = 1001
    UNKNOWN_EXCEPTION = 1002


if compat.PY3:
    class Error(Exception):
        code = None

        def __init__(self, message=None):
            self.message = message

        def __str__(self):
            message = ' ' + self.message if self.message is not None else ''
            return 'Code: {}.{}'.format(self.code, message)

    class ServerException(Error):
        def __init__(self, message, code, nested=None):
            self.message = message
            self.code = code
            self.nested = nested

        def __str__(self):
            nested = '\nNested: {}'.format(self.nested) if self.nested else ''
            return 'Code: {}.{}\n{}'.format(self.code, nested, self.message)

else:
    class Error(Exception):
        code = None

        def __init__(self, message=None):
            self.message = message

        def __unicode__(self):
            message = ' ' + self.message if self.message is not None else ''
            return 'Code: {}.{}'.format(self.code, message)

        def __str__(self):
            return compat.text_type(self).encode('utf-8')

    class ServerException(Error):
        def __init__(self, message, code, nested=None):
            self.message = message
            self.code = code
            self.nested = nested

        def __unicode__(self):
            nested = '\nNested: {}'.format(self.nested) if self.nested else ''
            return 'Code: {}.{}\n{}'.format(self.code, nested, self.message)

        def __str__(self):
            return compat.text_type(self).encode('utf-8')


class LogicalError(Error):
    code = ErrorCodes.LOGICAL_ERROR


class UnknownTypeError(Error):
    code = ErrorCodes.UNKNOWN_TYPE


class ChecksumDoesntMatchError(Error):
    code = ErrorCodes.CHECKSUM_DOESNT_MATCH


class TypeMismatchError(Error):
    code = ErrorCodes.TYPE_MISMATCH


class UnknownCompressionMethod(Error):
    code = ErrorCodes.UNKNOWN_COMPRESSION_METHOD


class TooLargeStringSize(Error):
    code = ErrorCodes.TOO_LARGE_STRING_SIZE


class NetworkError(Error):
    code = ErrorCodes.NETWORK_ERROR


class SocketTimeoutError(Error):
    code = ErrorCodes.SOCKET_TIMEOUT


class UnexpectedPacketFromServerError(Error):
    code = ErrorCodes.UNEXPECTED_PACKET_FROM_SERVER


class UnknownPacketFromServerError(Error):
    code = ErrorCodes.UNKNOWN_PACKET_FROM_SERVER


class CannotParseUuidError(Error):
    code = ErrorCodes.CANNOT_PARSE_UUID



from datetime import date, datetime
from enum import Enum
from uuid import UUID

from .compat import text_type, string_types


escape_chars_map = {
    "\b": "\\b",
    "\f": "\\f",
    "\r": "\\r",
    "\n": "\\n",
    "\t": "\\t",
    "\0": "\\0",
    "\a": "\\a",
    "\v": "\\v",
    "\\": "\\\\",
    "'": "\\'"
}


def escape_param(item):
    if item is None:
        return 'NULL'

    elif isinstance(item, datetime):
        return "'%s'" % item.strftime('%Y-%m-%d %H:%M:%S')

    elif isinstance(item, date):
        return "'%s'" % item.strftime('%Y-%m-%d')

    elif isinstance(item, string_types):
        return "'%s'" % ''.join(escape_chars_map.get(c, c) for c in item)

    elif isinstance(item, list):
        return "[%s]" % ', '.join(text_type(escape_param(x)) for x in item)

    elif isinstance(item, tuple):
        return "(%s)" % ', '.join(text_type(escape_param(x)) for x in item)

    elif isinstance(item, Enum):
        return item.value

    elif isinstance(item, UUID):
        return "'%s'" % str(item)

    else:
        return item


def escape_params(params):
    escaped = {}

    for key, value in params.items():
        escaped[key] = escape_param(value)

    return escaped



from datetime import date, timedelta

from .base import FormatColumn


class DateColumn(FormatColumn):
    ch_type = 'Date'
    py_types = (date, )
    format = 'H'

    epoch_start = date(1970, 1, 1)

    def before_write_item(self, value):
        return (value - self.epoch_start).days

    def after_read_item(self, value):
        return self.epoch_start + timedelta(value)



import os
import sys
import re
from codecs import open

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


PY34 = sys.version_info[0:2] >= (3, 4)

install_requires = []
if not PY34:
    install_requires.append('enum34')


def read_version():
    regexp = re.compile('^VERSION\W*=\W*\(([^\(\)]*)\)')
    init_py = os.path.join(here, 'src', '__init__.py')
    with open(init_py, encoding='utf-8') as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1).replace(', ', '.')
        else:
            raise RuntimeError('Cannot find version in src/__init__.py')


with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='clickhouse-driver',
    version=read_version(),

    description='Python driver with native interface for ClickHouse',
    long_description=long_description,

    url='https://github.com/mymarilyn/clickhouse-driver',

    author='Konstantin Lebedev',
    author_email='kostyan.lebedev@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',


        'Environment :: Console',


        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',


        'License :: OSI Approved :: MIT License',


        'Operating System :: OS Independent',


        'Programming Language :: SQL',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',

        'Topic :: Database',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],

    keywords='ClickHouse db database cloud analytics',

    packages=[
        p.replace('src', 'clickhouse_driver')
        for p in find_packages(exclude=['tests'])
        if p.startswith('src')
    ],
    package_dir={
        'clickhouse_driver': 'src',
    },
    install_requires=install_requires,
    extras_require={
        'lz4': ['lz4', 'clickhouse-cityhash==1.0.2.1'],
        'zstd': ['zstd', 'clickhouse-cityhash==1.0.2.1']
    },
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'mock',
        'freezegun',
        'lz4', 'zstd',
        'clickhouse-cityhash==1.0.2.1'
    ],
)



from . import errors, defines
from .block import Block
from .connection import Connection
from .protocol import ServerPacketTypes
from .util.escape import escape_params
from .util.helpers import chunks


class QueryResult(object):
    def __init__(self, with_column_types=False):
        self.with_column_types = with_column_types

        self.data = []
        self.columns_with_types = []

        super(QueryResult, self).__init__()

    def get_result(self):
        if self.with_column_types:
            return self.data, self.columns_with_types
        else:
            return self.data


class Progress(object):
    def __init__(self, client_connection, result, progress_gen):
        self.connection = client_connection
        self.result = result
        self.progress_gen = progress_gen

        super(Progress, self).__init__()

    def __iter__(self):
        return self

    def next(self):
        try:
            return next(self.progress_gen)
        except Exception as ex:
            if not isinstance(ex, StopIteration):
                self.connection.disconnect()
            raise

    # For Python 3.
    __next__ = next

    def get_result(self):
        # Read all progress packets.
        for _ in self:
            pass

        return self.result.get_result()


class Client(object):
    def __init__(self, *args, **kwargs):
        self.insert_block_size = kwargs.pop(
            'insert_block_size', defines.DEFAULT_INSERT_BLOCK_SIZE
        )
        self.connection = Connection(*args, **kwargs)
        super(Client, self).__init__()

    def disconnect(self):
        self.connection.disconnect()

    def receive_result(self, with_column_types=False, progress=False,
                       columnar=False):
        result = QueryResult(with_column_types=with_column_types)

        if progress:
            progress_gen = self.receive_progress_result(result, columnar)
            return Progress(self.connection, result, progress_gen)

        else:
            self.receive_no_progress_result(result, columnar)
            return result.get_result()

    def receive_progress_result(self, result, columnar=False):
        rows_read, approx_rows_to_read = 0, 0
        while True:
            packet = self.receive_packet()
            if not packet:
                break

            if packet is True:
                continue

            progress = getattr(packet, 'progress', None)
            if progress:
                if progress.new_total_rows:
                    approx_rows_to_read += progress.new_total_rows

                rows_read += progress.new_rows

                yield rows_read, approx_rows_to_read

            else:
                self.store_query_result(packet, result, columnar)

    def receive_no_progress_result(self, result, columnar=False):
        while True:
            packet = self.receive_packet()
            if not packet:
                break

            if packet is True:
                continue

            self.store_query_result(packet, result, columnar)

    def store_query_result(self, packet, result, columnar=False):
        block = getattr(packet, 'block', None)
        if block is None:
            return

        # Header block contains no rows. Pick columns from it.
        if block.rows:
            if columnar:
                columns = block.get_columns()
                if result.data:
                    # Extend corresponding column.
                    for i, column in enumerate(columns):
                        result.data[i] += column
                else:
                    result.data.extend(columns)
            else:
                result.data.extend(block.get_rows())

        elif not result.columns_with_types:
            result.columns_with_types = block.columns_with_types

    def receive_packet(self):
        packet = self.connection.receive_packet()

        if packet.type == ServerPacketTypes.EXCEPTION:
            raise packet.exception

        elif packet.type == ServerPacketTypes.PROGRESS:
            return packet

        elif packet.type == ServerPacketTypes.END_OF_STREAM:
            return False

        elif packet.type == ServerPacketTypes.DATA:
            return packet

        elif packet.type == ServerPacketTypes.TOTALS:
            return packet

        elif packet.type == ServerPacketTypes.EXTREMES:
            return packet

        else:
            return True

    def execute(self, query, params=None, with_column_types=False,
                external_tables=None, query_id=None, settings=None,
                types_check=False, columnar=False):
        self.connection.force_connect()

        try:
            # INSERT queries can use list or tuple of list/tuples/dicts.
            # For SELECT parameters can be passed in only in dict right now.
            is_insert = isinstance(params, (list, tuple))
            if is_insert:
                return self.process_insert_query(
                    query, params, external_tables=external_tables,
                    query_id=query_id, settings=settings,
                    types_check=types_check
                )
            else:
                return self.process_ordinary_query(
                    query, params=params, with_column_types=with_column_types,
                    external_tables=external_tables,
                    query_id=query_id, settings=settings,
                    types_check=types_check, columnar=columnar
                )

        except Exception:
            self.connection.disconnect()
            raise

    def execute_with_progress(
            self, query, params=None, with_column_types=False,
            external_tables=None, query_id=None, settings=None,
            types_check=False):

        self.connection.force_connect()

        try:
            return self.process_ordinary_query_with_progress(
                query, params=params, with_column_types=with_column_types,
                external_tables=external_tables,
                query_id=query_id, settings=settings, types_check=types_check
            )

        except Exception:
            self.connection.disconnect()
            raise

    def process_ordinary_query_with_progress(
            self, query, params=None, with_column_types=False,
            external_tables=None, query_id=None, settings=None,
            types_check=False, columnar=False):

        if params is not None:
            query = self.substitute_params(query, params)

        self.connection.send_query(query, query_id=query_id, settings=settings)
        self.connection.send_external_tables(external_tables,
                                             types_check=types_check)
        return self.receive_result(with_column_types=with_column_types,
                                   progress=True, columnar=columnar)

    def process_ordinary_query(
            self, query, params=None, with_column_types=False,
            external_tables=None, query_id=None, settings=None,
            types_check=False, columnar=False):

        if params is not None:
            query = self.substitute_params(query, params)

        self.connection.send_query(query, query_id=query_id, settings=settings)
        self.connection.send_external_tables(external_tables,
                                             types_check=types_check)
        return self.receive_result(with_column_types=with_column_types,
                                   columnar=columnar)

    def process_insert_query(self, query_without_data, data,
                             external_tables=None, query_id=None,
                             settings=None, types_check=False):
        self.connection.send_query(query_without_data, query_id=query_id,
                                   settings=settings)
        self.connection.send_external_tables(external_tables,
                                             types_check=types_check)

        sample_block = self.receive_sample_block()
        if sample_block:
            self.send_data(sample_block, data, types_check=types_check)
            packet = self.connection.receive_packet()
            if packet.exception:
                raise packet.exception

    def receive_sample_block(self):
        packet = self.connection.receive_packet()

        if packet.type == ServerPacketTypes.DATA:
            return packet.block

        elif packet.type == ServerPacketTypes.EXCEPTION:
            raise packet.exception

        else:
            message = self.connection.unexpected_packet_message('Data',
                                                                packet.type)
            raise errors.UnexpectedPacketFromServerError(message)

    def send_data(self, sample_block, data, types_check=False):
        for chunk in chunks(data, self.insert_block_size):
            block = Block(sample_block.columns_with_types, chunk,
                          types_check=types_check)
            self.connection.send_data(block)

        # Empty block means end of data.
        self.connection.send_data(Block())

    def cancel(self, with_column_types=False):
        # TODO: Add warning if already cancelled.
        self.connection.send_cancel()
        # Client must still read until END_OF_STREAM packet.
        return self.receive_result(with_column_types=with_column_types)

    def substitute_params(self, query, params):
        if not isinstance(params, dict):
            raise ValueError('Parameters are expected in dict form')

        escaped = escape_params(params)
        return query % escaped



from .reader import read_varint, read_binary_uint8


class BlockStreamProfileInfo(object):
    def __init__(self):
        self.rows = 0
        self.blocks = 0
        self.bytes = 0
        self.applied_limit = False  # bool
        self.rows_before_limit = 0
        self.calculated_rows_before_limit = 0  # bool

        super(BlockStreamProfileInfo, self).__init__()

    def read(self, fin):
        self.rows = read_varint(fin)
        self.blocks = read_varint(fin)
        self.bytes = read_varint(fin)
        self.applied_limit = bool(read_binary_uint8(fin))
        self.rows_before_limit = read_varint(fin)
        self.calculated_rows_before_limit = bool(read_binary_uint8(fin))
